/* X_window.c */
/* This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/* Routines to connect to the X server, and handle windows.
 */

/* Connect with X_init().  Disconnect with X_quit().  When connected, use
 * X_make_win() to create a window.  Call X_handle() to check for an event,
 * and handle it.  Call X_close_win() to close a window, or X_close_all()
 * to close all of them.  Call X_image() to set the image for the window.
 * Call X_name() to re-name a window.
 */


/* INCLUDES **************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "X_window.h"


/* LOCAL GLOBALS *********************/


/* Atoms for intercepting a close window command from the Window Manager. */
static Atom XA_WM_PROTOCOL=None,XA_WM_CLOSE=None;


/* For making XImage's.  Set by X_init(). */
static XVisualInfo visinfo;


/* My list of windows that this module keeps track of. */
static X_window_t *window_list=0;


/* EXPORTED GLOBALS ******************/


/* All of these globals are for reference only.
 * They are to be changed only by the subroutines in this module.
 */


/* The main display, the screen I'll use, and the root window.
 * Set by X_init().
 */
Display *X_display=0;
Screen *X_screen=0;
Window X_rootwin=None;


/* EXPORTED CODE *********************/


/* Finds and closes the window matching win.  Does nothing if that window
 * is not in the local list.
 */
void X_close_win(Window win){
	X_window_t *ptr,*prev=0;
	if(!X_display)return;
	for(ptr=window_list;ptr;prev=ptr,ptr=ptr->next){
		if(ptr->win==win){
			if(X_display){
				/* Got it.  Destroy, free, etc. */
				XDestroyWindow(X_display,win);
				/* DestroyImage frees ptr->image. */
				if(ptr->ximage)XDestroyImage(ptr->ximage);
			}
			if(!prev)window_list=ptr->next;else prev->next=ptr->next;
			free(ptr);return;
		}
	}
}


/* Finds and closes all the windows this module knows of. */
void X_close_all(){
	while(window_list)X_close_win(window_list->win);
}


/* Disconnects from the X server. */
void X_quit(){
	X_close_all();
  /* Close the connection. */
  if(X_display){XCloseDisplay(X_display);X_display=0;X_screen=0;X_rootwin=None;}
}


/* Connects to the X server.
 * Returns 0 for success, -1 for error (sets errno).
 */
int X_init(){
  /* Check if perhaps this has already been called. */
  if(X_display){errno=EEXIST;return -1;}
  /* Connect to the display, and get the default screen and root window. */
  X_display=XOpenDisplay(0);if(!X_display){
    printf("X_init():  Could not connect to X server.\n");
		errno=ENOENT;return -1;
  }
  X_screen=XDefaultScreenOfDisplay(X_display);if(!X_screen){
    printf("X_init():  Could not get default screen.\n");
		errno=ENOENT;return -1;
  }
  X_rootwin=XDefaultRootWindow(X_display);if(X_rootwin==None){
    printf("X_init():  Could not get default root window.\n");
		errno=ENOENT;return -1;
  }
  /* Set up stuff for images that I can work with. */
  if(!XMatchVisualInfo(X_display,0,24,TrueColor,&visinfo)){
    printf("X_init():  Cannot get a visual match.\n");
		errno=EDOM;return -1;
  }
	/* Get Atoms for detecting a close window command from the WM. */
	XA_WM_PROTOCOL=XInternAtom(X_display,"WM_PROTOCOLS",True);
	XA_WM_CLOSE=XInternAtom(X_display,"WM_DELETE_WINDOW",True);
	return 0;
}


/* Attempts to create a window.
 * x,y, width, and height are required, everything else may be 0 to not use
 * that feature.  name is just a name used for the window (like in title
 * bars).  key is a subroutine to call when it gets a key press or
 * something.
 * NOTE:  If x<=-width or y<=-height, this will skip using the x and y
 *        positions.
 * The four arguments to key are the window that got the event,
 * the closest I can get to a scancode from X, and pointers to the actual
 * event and my own window structure, just in case that would be helpful.
 * Do not make any changes to the X_window_t data.
 * goaway is a subroutine to call when the window manager tries to close
 * the window (like when the user clicks on the close window button).  If
 * goaway is null, then X_handle() will close the window.  Otherwise, it
 * will just call goaway(), and expect it to close the window.  goaway()
 * may call X_quit().
 * Returns the ID of the window (None in case of failure, and sets errno).
 */
Window X_make_win(int x,int y,unsigned width,unsigned height,char *name,
		void (*key)(Window win,unsigned code,const XEvent *event,const X_window_t *Xwin),
		void (*goaway)(Window win)){
	X_window_t *ptr;
	Window window;
  unsigned long mask;
  XSetWindowAttributes setwinattr;
  XTextProperty textprop;
  XSizeHints sizehints;
  XGCValues xgcvalues;
	if(!X_display){errno=ENOENT;return None;}
	/* First, try to allocate what I want. */
	ptr=(X_window_t *)malloc(sizeof(*ptr));
	if(!ptr){errno=ENOMEM;return None;}
  /* Set up a window with no background, that is black. */
  setwinattr.background_pixmap=None;
  setwinattr.background_pixel=0;
	setwinattr.bit_gravity=ForgetGravity;
	setwinattr.win_gravity=ForgetGravity;
  /* The events I want to receive -- key presses, exposure, and Destroy. */
  setwinattr.event_mask=KeyPressMask | KeyReleaseMask | ExposureMask |
		                    PointerMotionMask | ButtonPressMask |
												ButtonReleaseMask | StructureNotifyMask |
												CWBitGravity | CWWinGravity;
  /* Which winattr elements I set. */
  mask=CWBackPixmap | CWBackPixel | CWEventMask;
  window=XCreateWindow(X_display,X_rootwin,x,y,width,height,0,CopyFromParent,
	                     InputOutput,CopyFromParent,mask,&setwinattr);
  if(window==None){free(ptr);return None;}
  /* Get the window manager to deal with it.
   * Just picked XTextStyle.  Ignore errors.
	 */
	if(name){
		if(!XmbTextListToTextProperty(X_display,&name,1,XTextStyle,&textprop)){
			XSetWMName(X_display,window,&textprop);
			XFree(textprop.value);
		}
		if(!XmbTextListToTextProperty(X_display,&name,1,XTextStyle,&textprop)){
			XSetWMIconName(X_display,window,&textprop);
			XFree(textprop.value);
		}
	}
	/* Ignore errors telling the WM we want to intercept close requests, too.
	 * Also, request a certain position and size from the WM (since it can
	 * override the request in XCreateWindow()).
	 */
	if(XA_WM_PROTOCOL!=None){
		if(XA_WM_CLOSE!=None)XSetWMProtocols(X_display,window,&XA_WM_CLOSE,1);
	  sizehints.flags=PSize;
		if(x>-width && y>-height){
			sizehints.flags|=PPosition;
			sizehints.x=x;sizehints.y=y;
		}
		sizehints.width=width;sizehints.height=height;
		XSetWMNormalHints(X_display,window,&sizehints);
	}
	/* Setting mask to 0 means I don't actually use xgcvalues. */
  mask=0;ptr->gc=XCreateGC(X_display,window,mask,&xgcvalues);
	/* Insert it into my list. */
	ptr->win=window;ptr->ximage=0;ptr->key=key;ptr->goaway=goaway;
	ptr->xsize=ptr->ysize=ptr->x=ptr->y=0;ptr->image=0;
	ptr->cx=ptr->cy=-1;ptr->refresh=0;
	ptr->next=window_list;window_list=ptr;
  /* Put the window up. */
  XMapWindow(X_display,window);
	return window;
}


/* Refreshes the display of a window.
 * Returns 0 for success, -1 for error (sets errno).  I wouldn't consider
 * any errors from this to be too serious -- just try to continue.  Or,
 * perhaps, re-make the window.
 * Re-draws the whole window.
 */
int X_refresh(Window win){
	XWindowAttributes winattr;
	X_window_t *ptr;
	int width,height;
	if(!X_display){errno=ENOENT;return -1;}
	for(ptr=window_list;ptr;ptr=ptr->next){
		if(ptr->win==win){
			if(!XGetWindowAttributes(X_display,win,&winattr)){
				errno=EIO;return -1;
			}
			width=winattr.width;height=winattr.height;
			/* Sanity checks: */
			if(!ptr->ximage){
				ptr->xsize=ptr->ysize=ptr->x=ptr->y=0;
			}
			else{
				if(ptr->x<0)ptr->x=0;if(ptr->y<0)ptr->y=0;
				if(ptr->xsize<width)ptr->x=0;
				else if((ptr->x+width)>ptr->xsize)ptr->x=ptr->xsize-width;
				if(ptr->ysize<height)ptr->y=0;
				else if((ptr->y+height)>ptr->ysize)ptr->y=ptr->ysize-height;
				XPutImage(X_display,win,ptr->gc,ptr->ximage,ptr->x,ptr->y,0,0,
						ptr->xsize>width?width:ptr->xsize,ptr->ysize>height?height:ptr->ysize);
			}
			XClearArea(X_display,win,ptr->xsize,0,width-ptr->xsize,height,0);
			XClearArea(X_display,win,0,ptr->ysize,width,height-ptr->ysize,0);
			return 0;
		}
	}
	errno=EBADF;return -1;
}


/* Sets the image shown in a window.
 * Returns 0 for success, -1 for error (and sets errno).
 * image has the data, which is xs by ys, RGB for each pixel.
 * image may be NULL for no image.
 * Currently, this routine has an ugly hack -- it has to make a copy of
 * the image data to spread it out -- X wants 4 bytes per pixel (doesn't
 * use the fourth one).  The data is kept with the window, though, so that
 * you can change your image data, without affecting the image on the
 * screen.
 * On the plus side, that helps me with re-scaling.  If either xf or yf is
 * 0, no scaling is done -- the image is directly copied over as an xs by
 * ys image.  If neither xf or yf is 0, then the image is scaled using a
 * very simple method so that the image for the window is xf by yf.
 */
int X_image(Window win,char *image,unsigned short xs,unsigned short ys,
		unsigned short xf,unsigned short yf){
	unsigned soff,doff,sx,sy,dx,dy,xerr,yerr,s,scale;
	if(!X_display){errno=ENOENT;return -1;}
	if(!xf || !yf){scale=0;xf=xs;yf=ys;}else scale=1;
	X_window_t *ptr;
	for(ptr=window_list;ptr;ptr=ptr->next){
		if(ptr->win==win){
			/* Got to do this before I mess with the image.
			 * DestroyImage frees ptr->image.
			 */
			if(ptr->ximage){XDestroyImage(ptr->ximage);ptr->ximage=0;ptr->image=0;}
			if(!image){
				ptr->xsize=ptr->ysize=ptr->x=ptr->y=0;
				return X_refresh(win);
			}
			s=xf*yf*4;ptr->image=(char *)malloc(s);if(!ptr->image){
				ptr->xsize=ptr->ysize=ptr->x=ptr->y=0;
				X_refresh(win);errno=ENOMEM;return -1;
			}
			/* Below I specifically say I'm going to use RGB in X_image(). */
			if(!scale){
				/* Simple copying translation. */
				for(soff=0,doff=0;doff<s;soff+=3,doff+=4){
					ptr->image[doff+3]=image[soff+2];
					ptr->image[doff+2]=image[soff+1];
					ptr->image[doff+1]=image[soff  ];
					ptr->image[doff]=0;
				}
			}
			else{
				/* Scaling translation. */
				doff=0;for(yerr=sy=dy=0;dy<yf;++dy){
					soff=3*sy*xs;
					for(xerr=sx=dx=0;dx<xf;++dx,doff+=4){
						if(sx<xs && sy<ys){
							ptr->image[doff+3]=image[soff+2];
							ptr->image[doff+2]=image[soff+1];
							ptr->image[doff+1]=image[soff  ];
						}
						else memset((ptr->image)+doff,0,4);
						for(xerr+=xs;xerr>=xf;xerr-=xf,++sx,soff+=3);
					}
					for(yerr+=ys;yerr>=yf;yerr-=yf,++sy);
				}
			}
			ptr->ximage=XCreateImage(X_display,visinfo.visual,24,ZPixmap,0,
			                         ptr->image,xf,yf,8,xf*4);
			/* If that failed, well, I've done all I can.  Just refresh.
			 * That will help take care of ptr->x and ptr->y.
			 */
			if(ptr->ximage){
				/* This says I'm using RGB: */
				ptr->ximage->byte_order=MSBFirst;
				ptr->xsize=xf;ptr->ysize=yf;
			}
			else{ptr->xsize=ptr->ysize=ptr->x=ptr->y=0;}
			return X_refresh(win);
		}
	}
	errno=EBADF;return -1;
}


/* Allows you to reposition a window.  Makes requests to both X and the WM
 * (which tends to get priority).
 * Returns 0 for success, -1 for error (and sets errno).
 */
int X_reposition(Window win,int x,int y){
	int ret=0;
	unsigned int mask;
	XWindowChanges values;
  XSizeHints sizehints;
	X_window_t *ptr;
	if(!X_display){errno=ENOENT;return -1;}
	for(ptr=window_list;ptr;ptr=ptr->next){
		if(ptr->win==win){
			mask=CWX | CWY;values.x=x;values.y=y;
			ret=XConfigureWindow(X_display,win,mask,&values);
			if(ret){errno=EIO;return -1;}
			/* That was talking to X -- now the WM if I think there is one. */
			if(XA_WM_PROTOCOL!=None){
			  sizehints.flags=PPosition;sizehints.x=x;sizehints.y=y;
				XSetWMNormalHints(X_display,win,&sizehints);
			}
			return X_refresh(win);
		}
	}
	errno=EBADF;return -1;
}


/* Allows you to resize a window.  Makes requests to both X and the WM
 * (which tends to get priority).
 * Negative values for width and height are replaced by the size of the
 * image in the window, if there is one, or other defaults otherwise.
 * Returns 0 for success, -1 for error (and sets errno).
 */
int X_resize(Window win,int width,int height){
	int ret=0;
	unsigned int mask;
	XWindowChanges values;
  XSizeHints sizehints;
	X_window_t *ptr;
	if(!X_display){errno=ENOENT;return -1;}
	for(ptr=window_list;ptr;ptr=ptr->next){
		if(ptr->win==win){
			mask=CWWidth | CWHeight;
			if(width<0){
				if(ptr->ximage)width=ptr->xsize;
				else width=100;
			}
			values.width=width;
			if(height<0){
				if(ptr->ximage)height=ptr->ysize;
				else height=100;
			}
			values.height=height;
			ret=XConfigureWindow(X_display,win,mask,&values);
			if(ret){errno=EIO;return -1;}
			/* That was talking to X -- now the WM if I think there is one. */
			if(XA_WM_PROTOCOL!=None){
			  sizehints.flags=PSize;
				sizehints.width=width;sizehints.height=height;
				XSetWMNormalHints(X_display,win,&sizehints);
			}
			return X_refresh(win);
		}
	}
	errno=EBADF;return -1;
}


/* Allows you to re-name a window -- same name as in X_make_win().
 * Simply returns 0 if name is null.
 * Returns 0 for success, -1 for error (sets errno).
 * In the case of an error, I don't know what the window's name will show.
 */
int X_name(Window win,char *name){
  XTextProperty textprop;
	X_window_t *ptr;
	int ret=0;
	if(!X_display){errno=ENOENT;return -1;}
	if(!name)return 0;
	for(ptr=window_list;ptr;ptr=ptr->next){
		if(ptr->win==win){
			if(!(ret=XmbTextListToTextProperty(X_display,&name,1,XTextStyle,&textprop))){
				XSetWMName(X_display,win,&textprop);
				XFree(textprop.value);
			}
			if(ret){errno=EIO;return -1;}
			if(!(ret=XmbTextListToTextProperty(X_display,&name,1,XTextStyle,&textprop))){
				XSetWMIconName(X_display,win,&textprop);
				XFree(textprop.value);
			}
			if(ret){errno=EIO;return -1;}
			return 0;
		}
	}
	errno=EBADF;return -1;
}


/* Attempts to handle all the X events in the queue.  Returns immediately
 * if there are no events in the queue, unless nonblock is not set, in
 * which case it waits for one.  Will return after handling max number of
 * events. May call the key routine for windows.
 * There is a speed increase in letting it handle more than one at once,
 * since there may be a number of events that require a window to be
 * refreshed, and this routine will only refresh the windows at the end.
 * Returns number of events handled, normally.
 * Returns -1 (and sets errno to ENOENT) if the X display isn't initialized.
 */
int X_handle(int nonblock,int max){
	static X_window_t *ptr;
	static XEvent event;
	static KeySym *keysym=0;
	static unsigned code,length;
	static int x,y;
	int num=0;
	if(!X_display){errno=ENOENT;return -1;}
	if(nonblock && !XPending(X_display))return 0;
	XNextEvent(X_display,&event);for(;;){
		/* Find the window. */
		for(ptr=window_list;ptr;ptr=ptr->next){
			if(ptr->win==event.xany.window){
				switch(event.type){
					case KeyPress:
					case KeyRelease:
						keysym=XGetKeyboardMapping(X_display,event.xkey.keycode,1,(int *)&length);
						code=0;if(length>0)if(keysym[0]>=0 && keysym[0]<256)code=keysym[0];
						XFree(keysym);keysym=0;
						/* In case the key routine closes the window, or something,
						 * after this is called, exit immediately.
						 */
						if(ptr->key)ptr->key(ptr->win,code,&event,ptr);
						/* In case X_quit() was called, return. */
						if(!X_display)return num;
						break;
					case Expose:
						/* Refresh the new spot. */
						XClearArea(X_display,ptr->win,event.xexpose.x,event.xexpose.y,
								event.xexpose.width,event.xexpose.height,0);
						x=event.xexpose.x+ptr->x;y=event.xexpose.y+ptr->y;
						if(ptr->ximage)XPutImage(X_display,ptr->win,ptr->gc,ptr->ximage,
								x,y,event.xexpose.x,event.xexpose.y,
								event.xexpose.width<ptr->xsize-x?event.xexpose.width:ptr->xsize-x,
								event.xexpose.height<ptr->ysize-y?event.xexpose.height:ptr->ysize-y);
						break;
					case MotionNotify:
						/* Handle moving the image if the button is pressed. */
						if(!ptr->ximage)break;
						if(ptr->cx<0 || ptr->cy<0){
							ptr->cx=event.xmotion.x;ptr->cy=event.xmotion.y;
							break;
						}
						x=ptr->cx-event.xmotion.x;y=ptr->cy-event.xmotion.y;
						ptr->cx=event.xmotion.x;ptr->cy=event.xmotion.y;
						if(!(event.xmotion.state&Button1Mask))break;
						/* Let X_refresh() handle fixing ptr->x and ptr->y. */
						ptr->x+=x;ptr->y+=y;ptr->refresh=1;
						break;
					case ClientMessage:
						if(event.xclient.message_type==XA_WM_PROTOCOL &&
								event.xclient.data.l[0]==XA_WM_CLOSE){
							if(ptr->goaway)ptr->goaway(ptr->win);
							else X_close_win(ptr->win);
							/* In case X_quit() was called, return. */
							if(!X_display)return num;
						}
						break;
					case DestroyNotify:
						/* Much like the WM close message above. */
						if(ptr->goaway)ptr->goaway(ptr->win);
						else X_close_win(ptr->win);
						/* In case X_quit() was called, return. */
						if(!X_display)return num;
						break;
					default:break;
				}
				/* Handled the right window.  Now, get out of the window loop. */
				break;
			}
		}	/* The loop that searches for the right window. */
		/* Get the next event, if there is one. */
		++num;if(num>=max || !XPending(X_display))break;
		XNextEvent(X_display,&event);
	}
	/* Refresh all the windows that need it. */
	for(ptr=window_list;ptr;ptr=ptr->next){
		if(ptr->refresh){
			ptr->refresh=0;X_refresh(ptr->win);
		}
	}
	return num;
}
