/* X_window.c */

/* Routines to connect to the X server, and handle windows.
 */

/* Connect with X_init().  Disconnect with X_quit().  When connected, use
 * X_make_win() to create a window.  Call X_handle() to check for an event,
 * and handle it.  Call X_close_win() to close a window, or X_close_all()
 * to close all of them.  Call X_image() to set the image for the window.
 * Call X_name() to re-name a window.
 */


#ifndef X_WINDOW_H
#define X_WINDOW_H


/* INCLUDES **************************/


#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <errno.h>


/* TYPES *****************************/


/* This module maintains a linked list of the windows it controls.
 * These structures contain the data for the list.
 */
typedef struct X_window_s {
	Window win;                   /* ID */
	GC gc;                        /* For XImage's. */
	XImage *ximage;               /* The picture that goes in it. */
	char *image;                  /* Image data, in X format. */
	unsigned short xsize,ysize;   /* Size of the image. */
	short x,y;                    /* Where I'm looking in the image. */
	void (*key)(Window,unsigned,const XEvent *,const struct X_window_s *);
	                              /* Subroutine to call for handling keystrokes. */
	void (*goaway)(Window);       /* Subroutine for closing the window. */
	int cx,cy;                    /* Last seen cursor position. */
	                              /* Use -1 for unknown. */
	char refresh;                 /* Flag used by X_handle(). */
	struct X_window_s *next;
} X_window_t;


/* EXPORTED GLOBALS ******************/


/* All of these globals are for reference only.
 * They are to be changed only by the subroutines in this module.
 */


/* The main display, the screen I'll use, and the root window.
 * Set by X_init().
 */
Display *X_display;
Screen *X_screen;
Window X_rootwin;


/* PROTOTYPES ************************/


/* Finds and closes the window matching win.  Does nothing if that window
 * is not in the local list.
 */
void X_close_win(Window win);


/* Finds and closes all the windows this module knows of. */
void X_close_all();


/* Disconnects from the X server. */
void X_quit();


/* Connects to the X server.
 * Returns 0 for success, -1 for error (sets errno).
 */
int X_init();


/* Attempts to create a window.
 * x,y, width, and height are required, everything else may be 0 to not use
 * that feature.  name is just a name used for the window (like in title
 * bars).  key is a subroutine to call when it gets a key press or
 * something.
 * NOTE:  If x<=-width or y<=-height, this will skip using the x and y
 *        positions.
 * The four arguments to key are the window that got the event,
 * the closest I can get to a scancode from X, and pointers to the actual
 * event and my own window structure, just in case that would be helpful.
 * Do not make any changes to the X_window_t data.
 * goaway is a subroutine to call when the window manager tries to close
 * the window (like when the user clicks on the close window button).  If
 * goaway is null, then X_handle() will close the window.  Otherwise, it
 * will just call goaway(), and expect it to close the window.  goaway()
 * may call X_quit().
 * Returns the ID of the window (None in case of failure, and sets errno).
 */
Window X_make_win(int x,int y,unsigned width,unsigned height,char *name,
		void (*key)(Window win,unsigned code,const XEvent *event,const X_window_t *Xwin),
		void (*goaway)(Window win));


/* Refreshes the display of a window.
 * Returns 0 for success, -1 for error (sets errno).  I wouldn't consider
 * any errors from this to be too serious -- just try to continue.  Or,
 * perhaps, re-make the window.
 * Re-draws the whole window.
 */
int X_refresh(Window win);


/* Sets the image shown in a window.
 * Returns 0 for success, -1 for error (and sets errno).
 * image has the data, which is xs by ys, RGB for each pixel.
 * image may be NULL for no image.
 * Currently, this routine has an ugly hack -- it has to make a copy of
 * the image data to spread it out -- X wants 4 bytes per pixel (doesn't
 * use the fourth one).  The data is kept with the window, though, so that
 * you can change your image data, without affecting the image on the
 * screen.
 * On the plus side, that helps me with re-scaling.  If either xf or yf is
 * 0, no scaling is done -- the image is directly copied over as an xs by
 * ys image.  If neither xf or yf is 0, then the image is scaled using a
 * very simple method so that the image for the window is xf by yf.
 */
int X_image(Window win,char *image,unsigned short xs,unsigned short ys,
		unsigned short xf,unsigned short yf);


/* Allows you to reposition a window.  Makes requests to both X and the WM
 * (which tends to get priority).
 * Returns 0 for success, -1 for error (and sets errno).
 */
int X_reposition(Window win,int x,int y);


/* Allows you to resize a window.  Makes requests to both X and the WM
 * (which tends to get priority).
 * Negative values for width and height are replaced by the size of the
 * image in the window, if there is one, or other defaults otherwise.
 * Returns 0 for success, -1 for error (and sets errno).
 */
int X_resize(Window win,int width,int height);


/* Allows you to re-name a window -- same name as in X_make_win().
 * Simply returns 0 if name is null.
 * Returns 0 for success, -1 for error (sets errno).
 * In the case of an error, I don't know what the window's name will show.
 */
int X_name(Window win,char *name);


/* Attempts to handle all the X events in the queue.  Returns immediately
 * if there are no events in the queue, unless nonblock is not set, in
 * which case it waits for one.  Will return after handling max number of
 * events. May call the key routine for windows.
 * There is a speed increase in letting it handle more than one at once,
 * since there may be a number of events that require a window to be
 * refreshed, and this routine will only refresh the windows at the end.
 * Returns number of events handled, normally.
 * Returns -1 (and sets errno to ENOENT) if the X display isn't initialized.
 */
int X_handle(int nonblock,int max);


#endif
