/* This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/* demos.c */

/* A collection of plugins.
 * This file is still separated into chunks -- an includes chunk, a
 * defines chunk, etc.  Each chunk is split into parts based on which
 * plugin they are for.  With the exception of the includes chunk, all
 * chunks have sections for every plugin.
 */


/* INCLUDES **************************/


/* Something to include from the main program -- might as well include from
 * here (plus, it includes the necessary files).
 */
#include "demos.h"

/* For memset(). */
#include <string.h>
/* For rand stuff. */
#include <stdlib.h>
#include <fftw3.h>
#include <math.h>
#include "../lut.h"

#include <ft2build.h>  
#include FT_FREETYPE_H 

#define LUTMAX 255
lut_t lut[LUTMAX+1];

/* DEFINES ***************************/


/* ALL PLUGINS: */
/* Some handy constants that may be useful to all plugins.
 */

/* These are for updating an x,y position based on a direction.
 * They give the amount to add to x and y respectively for the UP direction
 * and the RIGHT direction.
 */
#define DIRS 8
static const int UPX[DIRS]={0,1,1,1,0,-1,-1,-1};
static const int UPY[DIRS]={-1,-1,0,1,1,1,0,-1};
static const int RIGHTX[DIRS]={1,1,0,-1,-1,-1,0,1};
static const int RIGHTY[DIRS]={0,1,1,1,0,-1,-1,-1};

/* MATRIX PLUGIN: */
/* Max number of trails and limits for speed (see the mat_trail_t structure).
 */
#define MAT_TRAIL 16
#define MAT_SPEED 16
#define MAT_SLOW0  32
#define MAT_SLOW  56
/* Two versions of the matrix plugin: define this to use the one that
 * clears the screen.  The one that does not clear the screen cannot add
 * brightnesses -- trails overlap.
 */
//#define MAT_CLEAR

/* FIRE PLUGIN: */
/* Number of sources for the fire.
 */
#define FIRE_SOURCE 32

/* BLOB PLUGIN: */

/* FLASH PLUGIN: */
#define FLASH_SPEED 2

#define TEXT_SPEED 2

#define DEBUG_SPEED 1

#define INT_SOURCE 2

/* TYPES *****************************/


/* MATRIX PLUGIN: */
/* Structure containing trail information.
 */
typedef struct {
	/* Brightness of the trail (0 for no trail). */
	pel_t bright;
	/* Position and speed.  Each frame, cnt is incremented by speed.
	 * (speed is mat_data.speed scaled to between 1 and MAT_SPEED.)
	 * The trail is moved down each time cnt exceeds slow, so that the
	 * average speed (pixel per frame) is speed/slow.
	 * slow, when chosen, is always between MAT_SLOW0 and MAT_SLOW0+MAT_SLOW,
	 * inclusive.
	 */
	int x,y;
	unsigned slow,cnt;
} mat_trail_t;

/* FIRE PLUGIN: */

/* BLOB PLUGIN: */

/* sources for the interference plugin */
typedef struct {
	/* position */
	unsigned x,y;
	/* how fast the waves decay */
	double decay;
	double frequency;
	double phase;
	double amplitude;
} int_src_t;


/* GLOBALS ***************************/


/* MATRIX PLUGIN: */
/* The globals that the main thread can change.
 */
data_t mat_data={
	density:(unsigned)(-1)/2,
	speed:(unsigned)(-1)/2,
	dir:0,
	color0:{0x00,0xFF,0x00},
	fft_iris_ceil:(double)0.0
};

/* FIRE PLUGIN: */
/* The globals that the main thread can change.
 */
data_t fire_data={
	density:(unsigned)(-1)/2,
	speed:(unsigned)(-1)/2,
	dir:0,
	color0:{0xFF,0x66,0x00},
	fft_iris_ceil:(double)0.0
};

/* BLOB PLUGIN: */
data_t blob_data={
	speed:(unsigned)(-1)/2,
	density:(unsigned)(-1)/2,
	dir:0,
	color0:{0xFF,0xFF,0xFF},
	fft_iris_ceil:(double)0.0
};


/* FLASH PLUGIN: */
data_t flash_data={
	speed:(unsigned)(-1)/2,
	color0:{0xFF,0xFF,0xFF},
	fft_iris_ceil:(double)0.0
};


/* TEXT PLUGIN: */
data_t text_data={
	speed:(unsigned)(-1)/2,
	color0:{0xFF,0xFF,0xFF},
	fft_iris_ceil:(double)0.0
};


/* DEBUG PLUGIN: */
data_t debug_data={
	speed:(unsigned)(-1)/2,
	color0:{0xFF,0xFF,0xFF},
	fft_iris_ceil:(double)0.0
};

/* EVENODD PLUGIN: */
data_t evenodd_data={
	speed:(unsigned)(-1)/2,
	color0:{0xFF,0xFF,0xFF},
	fft_iris_ceil:(double)0.0,
	dir:0
};

/* INTERFERENCE PLUGIN: */
data_t int_data={
	speed:(unsigned)(-1)/2,
	density:(unsigned)(-1)/2,
	color0:{0xFF,0xFF,0xFF},
	fft_iris_ceil:(double)0.0
};

/* STATIC GLOBALS ********************/


/* ALL PLUGINS: */
/* Some plugins use a spare frame, so that they can base changes off of the
 * original image, and not an image that has been partially changed.  They
 * can use this spare frame, as long as they do not depend on it remaining
 * unchanged between calls:
 */
static pel_t spare_frame[GRIDSIZE];

/* MATRIX PLUGIN: */
/* The important one:
 * Stores all the trails -- statically allocated and initialized by
 * mat_init().
 */
static mat_trail_t mat_trail[MAT_TRAIL];

/* FIRE PLUGIN: */
/* Also, the location of every source.
 */
int fire_x[FIRE_SOURCE],fire_y[FIRE_SOURCE];

/* BLOB PLUGIN: */

/* TEXT PLUGIN: */

/* INTERFERENCE PLUGIN: */
static int_src_t int_src[INT_SOURCE];

/* FREETYPE ******************************************/

unsigned radius = 0;

unsigned xa,ya,debug_c;

/* DEBUG PLUGIN */

/* ROUTINES **************************/


/* MATRIX PLUGIN: */

int mat_init(rand_t seed){
	unsigned a;
	/* Init rand num generator and data. */
	srand(seed);
	for(a=0;a<MAT_TRAIL;++a)mat_trail[a].bright=0;
	return 0;
}

int mat_gen(pel_t *frame){
	static unsigned char draw;
	static unsigned a,b,c,o,dir,dim,speed;
	static int x,y,upx,upy;
	/* Sanity check: */
	if(!frame)return -1;
#ifdef MAT_CLEAR
	/* Clear the frame. */
	memset(frame,0,GRIDSIZE);
#endif
	/* Define some handy numbers.  Try to do it in such a way that it is
	 * independent of particular integer sizes.
	 */
	dir=mat_data.dir%DIRS;upx=UPX[dir];upy=UPY[dir];
	dim=(unsigned)(-1)/GRIDMAX;dim=mat_data.density/dim;
	if(dim>=GRIDMAX)dim=1;else dim=GRIDMAX-dim;
	speed=(unsigned)(-1)/MAT_SPEED;speed=mat_data.speed/speed;
	if(speed>MAT_SPEED)speed=MAT_SPEED;if(speed==0)speed=1;
	/* Start, draw, and update the trails. */
	for(a=0;a<MAT_TRAIL;++a){
		if(!mat_trail[a].bright){
			/* Start a new trail somewhere in the grid, and then back it up until
			 * it hits an edge.  Not very efficient, but not too shabby for the
			 * way it works for any direction.
			 */
			mat_trail[a].bright=rand()%(GRIDMAX+1);
			x=rand()%GRIDX;y=rand()%GRIDY;
			for(;x>0 && x<(GRIDX-1) && y>0 && y<(GRIDY-1);x+=upx,y+=upy);
			mat_trail[a].x=x;mat_trail[a].y=y;
			mat_trail[a].slow=rand()%(MAT_SLOW+1)+MAT_SLOW0;
			mat_trail[a].cnt=0;
		}
		/* Check this out: using defines for whether or not to compile
		 * comments!!:
		 */
#ifdef MAT_CLEAR
		/* Draw this trail.  For now, just write to the first byte of each pel,
		 * for adding brightnesses.
		 */
#endif
		b=mat_trail[a].bright;draw=0;
		for(x=mat_trail[a].x,y=mat_trail[a].y;;x+=upx,y+=upy){
			if(x>=0 && x<GRIDX && y>=0 && y<GRIDY){
				o=(y*GRIDX+x)*PEL_NUMEL;
#ifdef MAT_CLEAR
				c=(unsigned)frame[o];c+=b;if(c>GRIDMAX)c=GRIDMAX;
				frame[o]=c;
#else
				frame[o  ]=(unsigned)(mat_data.color0[0])*b/GRIDMAX;
				frame[o+1]=(unsigned)(mat_data.color0[1])*b/GRIDMAX;
				frame[o+2]=(unsigned)(mat_data.color0[2])*b/GRIDMAX;
				frame[o+3]=(unsigned)(mat_data.color0[3])*b/GRIDMAX;
#endif
				draw=1;
			}
			/* The grid is convex, so if we have drawn something, but are now off
			 * the grid, we are done.
			 */
			else if(draw)break;
			/* Don't quit until after I have drawn a blank -- that's the closest
			 * I have for erasing when I'm not clearing the grid.
			 */
			if(!b){
				/* If we never drew anything, this trail is gone past the edge --
				 * disable.
				 */
				if(!draw)mat_trail[a].bright=0;
				break;
			}
			if(b<dim)b=0;else b-=dim;
		}
		/* Update this trail.  The b count is just to prevent an infinite loop
		 * if something gets set to 0 that should not be.
		 */
		mat_trail[a].cnt+=speed;
		for(b=GRIDMAX;b>0 && mat_trail[a].cnt>=mat_trail[a].slow;){
			mat_trail[a].cnt-=mat_trail[a].slow;
			mat_trail[a].x-=upx;mat_trail[a].y-=upy;
		}
	} /* for(...) -- loop through all trails. */
#ifdef MAT_CLEAR
	/* Now, actually color the frame from the brightnesses I have stored.
	 */
	for(o=0;o<GRIDSIZE;){
		b=frame[o];
		frame[o++]=(unsigned)(mat_data.color0[0])*b/GRIDMAX;
		frame[o++]=(unsigned)(mat_data.color0[1])*b/GRIDMAX;
		frame[o++]=(unsigned)(mat_data.color0[2])*b/GRIDMAX;
		frame[o++]=(unsigned)(mat_data.color0[3])*b/GRIDMAX;
	}
#endif
	//fft_frame_iris(mat_data.fft_iris_ceil,frame);
	return 0;
}


/* FIRE PLUGIN: */

int fire_init(rand_t seed){
	unsigned a;
	/* Init. random number generator, and positions of sources. */
	srand(seed);
	for(a=0;a<FIRE_SOURCE;++a){fire_x[a]=rand()%GRIDX;fire_y[a]=rand()%GRIDY;}
	return 0;
}

int fire_gen(pel_t *frame){
	static unsigned a,b,c,A,R,G,B,o,o2,dir,dim,speed;
	static int x,y,upx,upy,rightx,righty;
	/* Sanity check: */
	if(!frame)return -1;

	/* Define some handy numbers.  Try to do it in such a way that it is
	 * independent of particular integer sizes.
	 */
	dir=fire_data.dir%DIRS;
	upx=UPX[dir];upy=UPY[dir];rightx=RIGHTX[dir];righty=RIGHTY[dir];
	dim=(unsigned)(-1)/GRIDMAX;dim=fire_data.density/dim;
	if(dim>GRIDMAX)dim=GRIDMAX;

	/* Add the sources first, so that the fire effect will blur them out.
	 */
	for(a=0;a<FIRE_SOURCE;++a){
		/* Sources brighten where they are. */
		/* Sanity check: */
		if(fire_x[a]<0)fire_x[a]=0;if(fire_x[a]>=GRIDX)fire_x[a]=GRIDX-1;
		if(fire_y[a]<0)fire_y[a]=0;if(fire_y[a]>=GRIDY)fire_y[a]=GRIDY-1;
		o=(fire_y[a]*GRIDX+fire_x[a])*PEL_NUMEL;
		memcpy(frame+o,fire_data.color0,sizeof(*frame)*PEL_NUMEL);
		/* Move them, randomly.  Do not worry about bounds -- I'll get them the
		 * next time around.
		 */
		/* FIXME: rand() may return a different type than speed. */
		if(rand()<(fire_data.speed)/4){
			b=rand()%DIRS;fire_x[a]+=UPX[b];fire_y[a]+=UPY[b];
		}
	}

	/* Back up the frame. */
	memcpy(spare_frame,frame,sizeof(spare_frame));

	/* Do the fire effect. */
	o=0;for(a=0;a<GRIDY;++a)for(b=0;b<GRIDX;++b,o+=PEL_NUMEL){
		/* Sum up the colors for the three pixels "below", with skewed weights.
		 */
		A=R=G=B=c=0;y=a-upy-righty;x=b-upx-rightx;
		if(x>=0 && x<GRIDX && y>=0 && y<GRIDY){
			c+=2;o2=(y*GRIDX+x)*PEL_NUMEL;
			A+=2*(unsigned)spare_frame[o2  ];
			R+=2*(unsigned)spare_frame[o2+1];
			G+=2*(unsigned)spare_frame[o2+2];
			B+=2*(unsigned)spare_frame[o2+3];
		}
		y+=righty;x+=rightx;
		if(x>=0 && x<GRIDX && y>=0 && y<GRIDY){
			c+=4;o2=(y*GRIDX+x)*PEL_NUMEL;
			A+=4*(unsigned)spare_frame[o2  ];
			R+=4*(unsigned)spare_frame[o2+1];
			G+=4*(unsigned)spare_frame[o2+2];
			B+=4*(unsigned)spare_frame[o2+3];
		}
		y+=righty;x+=rightx;
		if(x>=0 && x<GRIDX && y>=0 && y<GRIDY){
			c+=2;o2=(y*GRIDX+x)*PEL_NUMEL;
			A+=2*(unsigned)spare_frame[o2  ];
			R+=2*(unsigned)spare_frame[o2+1];
			G+=2*(unsigned)spare_frame[o2+2];
			B+=2*(unsigned)spare_frame[o2+3];
		}
		/* Add up the current row. */
		y=a-righty;x=b-rightx;
		if(x>=0 && x<GRIDX && y>=0 && y<GRIDY){
			++c;o2=(y*GRIDX+x)*PEL_NUMEL;
			A+=(unsigned)spare_frame[o2  ];
			R+=(unsigned)spare_frame[o2+1];
			G+=(unsigned)spare_frame[o2+2];
			B+=(unsigned)spare_frame[o2+3];
		}
		y+=righty;x+=rightx;
		if(x>=0 && x<GRIDX && y>=0 && y<GRIDY){
			++c;o2=(y*GRIDX+x)*PEL_NUMEL;
			A+=(unsigned)spare_frame[o2  ];
			R+=(unsigned)spare_frame[o2+1];
			G+=(unsigned)spare_frame[o2+2];
			B+=(unsigned)spare_frame[o2+3];
		}
		y+=righty;x+=rightx;
		if(x>=0 && x<GRIDX && y>=0 && y<GRIDY){
			++c;o2=(y*GRIDX+x)*PEL_NUMEL;
			A+=(unsigned)spare_frame[o2  ];
			R+=(unsigned)spare_frame[o2+1];
			G+=(unsigned)spare_frame[o2+2];
			B+=(unsigned)spare_frame[o2+3];
		}
		/* Add the current color (again, just to make sure c is not 0). */
		c+=2;A+=2*(unsigned)frame[o];R+=2*(unsigned)frame[o+1];
		G+=2*(unsigned)frame[o+2];B+=2*(unsigned)frame[o+3];
		/* Average and attenuate (the above ensures that c cannot be 0). */
		A/=c;R/=c;G/=c;B/=c;
		A*=dim;A/=GRIDMAX;R*=dim;R/=GRIDMAX;G*=dim;G/=GRIDMAX;B*=dim;B/=GRIDMAX;
		/* Set the color. */
		frame[o]=A;frame[o+1]=R;frame[o+2]=G;frame[o+3]=B;
	}
	//fft_frame_iris(flash_data.fft_iris_ceil,frame);
	return 0;
}


/* BLOB PLUGIN: */
int blob_gen(pel_t *frame){
	int x,y,x2,y2,diff;
	unsigned a,o,o2,blobcount,dens,speed,speed2;

	/* Define some handy numbers.  Try to do it in such a way that it is
	 * independent of particular integer sizes.
	 */
	dens=(unsigned)(-1)/GRIDSIZE;dens=blob_data.density/dens;
	if(dens>GRIDSIZE)dens=GRIDSIZE;
	/* The scale-down factors here need to more than compensate for the
	 * scale-up that may come from adding up the weights.  See SCALING ISSUE
	 * below.
	 */
	speed=blob_data.speed/32;speed2=0;

	/* First, go through frame and find every pel that is close to the right
	 * color.  Save that in the 0 channel of spare_frame.
	 */
	blobcount=0;for(o=0;o<GRIDSIZE;o+=PEL_NUMEL){
		x =(int)(unsigned)frame[o  ]-(int)(unsigned)blob_data.color0[0]*rand();
		x+=(int)(unsigned)frame[o+1]-(int)(unsigned)blob_data.color0[1]*rand();
		x+=(int)(unsigned)frame[o+2]-(int)(unsigned)blob_data.color0[2]*rand();
		x+=(int)(unsigned)frame[o+3]-(int)(unsigned)blob_data.color0[3]*rand();
		if(x<GRIDMAX/16 && x>-GRIDMAX/16){
			spare_frame[o]=1;++blobcount;
		}else spare_frame[o]=0;
	}
	/* If nothing was found, create something. */
	if(!blobcount){
		x=(unsigned)rand()%GRIDX;y=(unsigned)rand()%GRIDY;
		o=(y*GRIDX+x)*PEL_NUMEL;spare_frame[o]=1;++blobcount;
	}
	/* Now, go through spare_frame and store the number of neighboring pels
	 * that are blob-like.  The 1 channel is higher for pels that have lots
	 * of neighbors, and the 2 channel is the reverse.  Channel 1 is used in
	 * deciding what to turn into blobs, and channel 2 is for removing blobs.
	 */
	o=0;for(y=0;y<GRIDY;++y)for(x=0;x<GRIDX;++x,o+=PEL_NUMEL){
		spare_frame[o+1]=spare_frame[o+2]=0;
		/* Check all neighbors.  Wrap around for coolness.
		 * Give higher scores for directions close to dir.
		 * Cheat: skip diagonals.
		 */
		for(a=0;a<DIRS;a+=2){
			diff=(int)a-(int)blob_data.dir;if(diff<0)diff+=DIRS;
			x2=x+UPX[a];y2=y+UPY[a];
			if(x2<0)x2=GRIDX-1;if(x2>=GRIDX)x2=0;
			if(y2<0)y2=GRIDY-1;if(y2>=GRIDY)y2=0;
			o2=(y2*GRIDX+x2)*3;
			if(spare_frame[o2]){
				if(diff>2 && diff<DIRS-2)spare_frame[o+1]+=4;
				else spare_frame[o+1]+=1;
			}else{
				if(diff>2 && diff<DIRS-2)spare_frame[o+2]+=4;
				else spare_frame[o+2]+=1;
			}
		}
	}
	/* Non-blobs with a high score have a chance at becoming blobs.
	 * Blobs with a low score have a chance at becoming non-blobs.
	 * The likelihoods are weighted to keep blobcount (number of blobs) at
	 * the ratio of the grid specified by density.
	 * Base this on the value of blobcount BEFORE changing any of these.
	 */
	for(o=0;o<GRIDSIZE;o+=PEL_NUMEL){
		if(spare_frame[o]==1){
			/* SCALING ISSUE:  speed must be scaled down enough to not wrap here. */
			if(rand()<spare_frame[o+2]*(blobcount>=dens?speed:speed2)){
				/* Be sure to mark for erasure: */
				spare_frame[o]=2;
			}
		}else{
			/* SCALING ISSUE:  speed must be scaled down enough to not wrap here. */
			if(rand()<spare_frame[o+1]*(blobcount<dens?speed:speed2)){
				spare_frame[o]=1;
			}
		}
	}
	/* Now, update frame. */
	for(o=0;o<GRIDSIZE;o+=PEL_NUMEL){
		if(spare_frame[o]==1)memcpy(frame+o,blob_data.color0,sizeof(blob_data.color0));
		if(spare_frame[o]==2)memset(frame+o,0,PEL_NUMEL*sizeof(*frame));
	}
	//fft_frame_iris(blob_data.fft_iris_ceil,frame);
	return 0;
}

/* FLASH PLUGIN: 
 * Picks a random color and fades that color based upon speed */
int flash_gen(pel_t *frame){
	unsigned o,r,g,b,a,x,y,i,j;
	uint8_t audio_buff2[AUDIO_BUFFSIZE];
	uint8_t audio_buff3[AUDIO_BUFFSIZE];
	memcpy(audio_buff2,audio_buff,sizeof(audio_buff));
	memcpy(spare_frame,frame,sizeof(spare_frame));

	/* first thing is to do a fft on our audio_buf */
	double fft_in[AUDIO_BUFFSIZE/2];
	fftw_complex fft_out[AUDIO_BUFFSIZE/4];

	//fft_in=(double*)fftw_malloc(sizeof(double)*AUDIO_BUFFSIZE/2);
	//fft_out=fftw_malloc(sizeof(fftw_complex)*257);
	
	/* fill fft_in with the audio data */
	for(i=0,j=0;i<AUDIO_BUFFSIZE/2;i++,j+=2){
		fft_in[i]=labs(((audio_buff2[j]+audio_buff2[j+1])/2)-128)*4;
		//printf("%f\n",fft_in[i]);
	}

	/* execute the plan! */
	fftw_plan q_r2c;
	q_r2c=fftw_plan_dft_r2c_1d(AUDIO_BUFFSIZE/16,fft_in,fft_out,FFTW_ESTIMATE);
	fftw_execute(q_r2c);

	/* fill the thing with some data.  Why did I call this variable "huff"?
	 * I really regret it */

	double huff[GRIDX+1];
	double maxhuff;
	double avghuff;
	for(a=0;a<GRIDX;a++){
		huff[a]=powl(fft_out[a][0]*fft_out[a][0]+fft_out[a][1]*fft_out[a][1],0.5);
		//printf("%f\n",huff[a]);
	}

	/* find the max value */
	maxhuff=huff[1];
	for(a=2;a<GRIDX;a++){
		if(huff[a]>maxhuff){
			maxhuff=huff[a];
		}
	}
	avghuff=0;
	for(a=2;a<GRIDX;a++){
		avghuff+=huff[a];
	} avghuff=(avghuff/a)/maxhuff*2;
	if(avghuff!=avghuff)avghuff=0.0;
	//printf("%lf\n",avghuff);


	/* let's do this with our NOT fft'd thing.  This will give us our average
	 * value */
	double level = 0;
	for(a=0;a<AUDIO_BUFFSIZE/2;a++){
		level+=(audio_buff2[a]+audio_buff2[a+1])/2;
	} level=(level/a);
	//printf("%f\n",fabs(level-128));

	static unsigned speed;
	/* copy the frame into our spare */
	
	/* got this from the matrix code, basically normalizes the speed */
	speed=(unsigned)(-1)/FLASH_SPEED;speed=flash_data.speed/speed;
	if(speed>FLASH_SPEED)speed=FLASH_SPEED;if(speed==0)speed=1;

	/* pick a random color just in case */
	r=(unsigned)(rand()%255);
	g=(unsigned)(rand()%255);
	b=(unsigned)(rand()%255);
	a=(unsigned)(rand()%255);
	//printf("%lf\n",avghuff);
	
	for(o=0;o<GRIDSIZE;o+=PEL_NUMEL){
		/* each call sees frame[] decremented by the speed */
		if(spare_frame[o  ]>speed){ frame[o  ]=spare_frame[o  ]-speed; }
		if(spare_frame[o+1]>speed){ frame[o+1]=spare_frame[o+1]-speed; }
		if(spare_frame[o+2]>speed){ frame[o+2]=spare_frame[o+2]-speed; }
		if(spare_frame[o+3]>speed){ frame[o+3]=spare_frame[o+3]-speed; }
	
		/* if they're all less than speed, time to pick a new color for that
		 * frame */
		frame[o]=((level-128)+spare_frame[o])/1.5;
		frame[o+1]=((level-128)+spare_frame[o])/1.5;
		frame[o+2]=((level-128)+spare_frame[o])/1.5;
		frame[o+3]=((level-128)+spare_frame[o])/1.5;

		if(avghuff>0.74){
				frame[o  ]=r;
				frame[o+1]=g;
				frame[o+2]=b;
				frame[o+3]=a;

			}
	}
	fftw_destroy_plan(q_r2c);
	//fftw_free(fft_out);
	//free(fft_in); 

	//fft_frame_iris(flash_data.fft_iris_ceil,frame);
	return 0;
}

int text_init(rand_t seed){
}

int text_gen(pel_t *frame){
	unsigned o,r,g,b,a,x,y,speed,i;
	speed=(unsigned)(-1)/TEXT_SPEED;speed=flash_data.speed/speed;
	if(speed>TEXT_SPEED)speed=TEXT_SPEED;if(speed==0)speed=1;
	for(i=0;i<speed;i++){

	/* this will hold points on a circle */
	o=0;for(y=0;y<GRIDY;++y)for(x=0;x<GRIDX;++x,o+=PEL_NUMEL){
		if((x-GRIDX/2)*(x-GRIDX/2)+(y-GRIDY/2)*(y-GRIDY/2)==radius){
		frame[o  ]=(unsigned)(text_data.color0[0]);
		frame[o+1]=(unsigned)(text_data.color0[1]);
		frame[o+2]=(unsigned)(text_data.color0[2]);
		frame[o+3]=(unsigned)(text_data.color0[3]);
			}
		else{ }
	}
	if(radius>GRIDX*10 && radius>GRIDY*10){ radius=0; }
	else{ radius++; }
	}
	
	//fft_frame_iris(text_data.fft_iris_ceil,frame);
	return 0;

}
int debug_init(rand_t seed){
	xa=ya=0;
	debug_c=0;
}


int debug_gen(pel_t *frame){
	unsigned o,r,g,b,a,x,y,i,speed;
	memcpy(spare_frame,frame,sizeof(spare_frame));

speed=(unsigned)(-1)/DEBUG_SPEED;speed=debug_data.speed/speed;
	if(speed>DEBUG_SPEED)speed=DEBUG_SPEED;if(speed==0)speed=1;
	for(i=0;i<speed;i++){

	o=0;for(y=0;y<GRIDY;++y)for(x=0;x<GRIDX;++x,o+=PEL_NUMEL){
		if(y==ya && x==xa){
			frame[o  ]=debug_data.color0[0];
			frame[o+1]=debug_data.color0[1];
			frame[o+2]=debug_data.color0[2];
			frame[o+3]=debug_data.color0[3];
		}
		else{
			frame[o]=spare_frame[o]/2;
			frame[o+1]=spare_frame[o+1]/2;
			frame[o+2]=spare_frame[o+2]/2;
			frame[o+3]=spare_frame[o+3]/2;
		}
	}
	xa++;
	if(xa==GRIDX){xa=0;ya++;}
	if(ya==GRIDY){ya=0;}
	debug_c++;
	//fft_frame_iris(debug_data.fft_iris_ceil,frame);
	}
	return 0;
}


int evenodd_init(rand_t seed){
}

/* number of EQ samples to make */
int evenodd_gen(pel_t *frame){
	unsigned o,r,g,b,a,x,y,i,j;
	uint8_t audio_buff2[AUDIO_BUFFSIZE];
	uint8_t audio_buff3[AUDIO_BUFFSIZE];
	memcpy(audio_buff2,audio_buff,sizeof(audio_buff));
	memcpy(spare_frame,frame,sizeof(spare_frame));

	/* first thing is to do a fft on our audio_buf */
	double fft_in[AUDIO_BUFFSIZE/2];
	fftw_complex fft_out[AUDIO_BUFFSIZE/4];

	//fft_in=(double*)fftw_malloc(sizeof(double)*AUDIO_BUFFSIZE/2);
	//fft_out=fftw_malloc(sizeof(fftw_complex)*257);
	
	/* fill fft_in with the audio data */
	for(i=0,j=0;i<AUDIO_BUFFSIZE/2;i++,j+=2){
		fft_in[i]=labs(((audio_buff2[j]+audio_buff2[j+1])/2)-128)*4;
		//printf("%f\n",fft_in[i]);
	}

	/* execute the plan! */
	fftw_plan q_r2c;
	q_r2c=fftw_plan_dft_r2c_1d(AUDIO_BUFFSIZE/16,fft_in,fft_out,FFTW_ESTIMATE);
	fftw_execute(q_r2c);

	/* fill the thing with some data.  Why did I call this variable "huff"?
	 * I really regret it */

	double huff[GRIDX+2];
	double maxhuff;
	double avghuff;
	for(a=0;a<GRIDX+1;a++){
		huff[a]=powl(fft_out[a][0]*fft_out[a][0]+fft_out[a][1]*fft_out[a][1],0.5);
		//printf("%f\n",huff[a]);
	}

	/* find the max value */
	maxhuff=huff[2];
	for(a=2;a<GRIDX+1;a++){
		if(huff[a]>maxhuff){
			maxhuff=huff[a];
		}
	}
	avghuff=0;
	for(a=2;a<GRIDX+1;a++){
		avghuff+=huff[a];
	} avghuff=(avghuff/a)/maxhuff*2;
	//printf("%lf\n",avghuff);

	/* now scale all the values */
	for(a=0;a<GRIDX+1;a++){
		huff[a]=huff[a]/maxhuff;
		//printf("%d: %d\n",a,(unsigned)(huff[a]*GRIDY));
	}


	/* let's do this with our NOT fft'd thing.  This will give us our average
	 * value */
	double level = 0;
	for(a=0;a<AUDIO_BUFFSIZE/2;a++){
		level+=(audio_buff2[a]+audio_buff2[a+1])/2;
	} level=(level/a);
	//printf("%f\n",fabs(level-128));

	o=0;a=0;for(y=0;y<GRIDY;y++)for(x=0;x<GRIDX;x++,o+=PEL_NUMEL,a++){
		switch (evenodd_data.dir){
			/* some wierd static */
			case 0:
				frame[o  ]=(fft_in[a]/0xFF)*evenodd_data.color0[0];
				frame[o+1]=(fft_in[a]/0xFF)*evenodd_data.color0[1];
				frame[o+2]=(fft_in[a]/0xFF)*evenodd_data.color0[2];
				frame[o+3]=(fft_in[a]/0xFF)*evenodd_data.color0[3];
				break;
			case 1:
					frame[o]=((level-128)+spare_frame[o])/1.5;
					frame[o+1]=((level-128)+spare_frame[o])/1.5;
					frame[o+2]=((level-128)+spare_frame[o])/1.5;
					frame[o+3]=((level-128)+spare_frame[o])/1.5;
					break;
			case 2:
				if(y<(unsigned)(avghuff*GRIDX)){
					frame[o]=evenodd_data.color0[0];
					frame[o+1]=evenodd_data.color0[1];
					frame[o+2]=evenodd_data.color0[2];
					frame[o+3]=evenodd_data.color0[3];
				}
				else{
					frame[o]=spare_frame[o]/2;
					frame[o+1]=spare_frame[o+1]/2;
					frame[o+2]=spare_frame[o+2]/2;
					frame[o+3]=spare_frame[o+3]/2;
				}
				break;
			case 3:
			if(y<(unsigned)(huff[x+1]*GRIDX)){
					frame[o]=evenodd_data.color0[0];
					frame[o+1]=evenodd_data.color0[1];
					frame[o+2]=evenodd_data.color0[2];
					frame[o+3]=evenodd_data.color0[3];
				}
				else{
					frame[o]=spare_frame[o]/2;
					frame[o+1]=spare_frame[o+1]/2;
					frame[o+2]=spare_frame[o+2]/2;
					frame[o+3]=spare_frame[o+3]/2;

				}
				break;
			default:
				break;
		}

	}

	fftw_destroy_plan(q_r2c);
	//fftw_free(fft_out);
	//free(fft_in); 
	return 0;
}

/* initialize our sources YYY*/	
int int_init(rand_t seed){
	unsigned a;
	srand(seed);
	for(a=0;a<INT_SOURCE;a++){
		/* [0,GRIDX] */
		int_src[a].x=GRIDX*((double)rand()/((double)(RAND_MAX)+(double)(1)));
		/* [0,GRIDY] */
		int_src[a].y=GRIDY*((double)rand()/((double)(RAND_MAX)+(double)(1)));
		/* [0,5] */
		int_src[a].decay=5.0*((double)rand()/((double)(RAND_MAX)+(double)(1)));
		/* [0,255] */
		int_src[a].frequency=0.25*((double)rand()/((double)(RAND_MAX)+(double)(1)));
		/* [0,2*PI] */
		int_src[a].phase=2.0*M_PI*((double)rand()/((double)(RAND_MAX)+(double)(1)));
		/* [0,1] */
		int_src[a].amplitude=((double)rand()/((double)(RAND_MAX)+(double)(1)));

  	lut_iter=LUTMAX;
	  lut_set(lut,  0,0x00,0x00,0xFF,0x00);
		lut_set(lut,LUTMAX,0x00,0x00,0xFF,0x00);
	  lut_set(lut, 32,0xFF,0xFF,0xFF,0xFF);
	  lut_set(lut, 64,0x00,0x00,0x00,0xDD);
	  lut_set(lut, 96,0xF0,0x60,0x00,0x60);
	  lut_set(lut,128,0x00,0xFF,0x00,0xFF);
	  lut_set(lut,160,0x00,0xFF,0x00,0x00);
	  lut_set(lut,192,0x00,0xFF,0x70,0x00);
	  lut_set(lut,224,0x00,0xFF,0xFF,0x00);
	  lut_fill(lut);

	}
	return 0;	
}

int int_gen(pel_t *frame){
	unsigned o,a,x,y,v;
	double t;
	double r;
	memset(frame,0,GRIDSIZE);
	memcpy(spare_frame,frame,sizeof(spare_frame));
	o=0;a=0;for(y=0;y<GRIDY;y++)for(x=0;x<GRIDX;x++,o+=PEL_NUMEL){

		for(a=0;a<INT_SOURCE;a++){
			r=pow(pow(abs(x-int_src[a].x),2.0)+pow(abs(y-int_src[a].y),2.0),0.5);
			v=(int_data.color0[0]+int_data.color0[1]+
			int_data.color0[2]+int_data.color0[3])/4;

			t=sin(int_src[a].frequency*r);
			frame[o  ]+=t*int_data.color0[0];
			frame[o+1]+=t*int_data.color0[1];
			frame[o+2]+=t*int_data.color0[2];
			frame[o+3]+=t*int_data.color0[3];
		}
	}


	/* walk around */
	for(a=0;a<INT_SOURCE;a++){
		/* walk x */
		if(((double)rand()/((double)(RAND_MAX)+(double)(1)))>0.5){
			if(int_src[a].x++>GRIDX)int_src[a].x=GRIDX;
		}
		else{
			if(int_src[a].x--==0)int_src[a].x=0;
		}
		/* walk y */
		if(((double)rand()/((double)(RAND_MAX)+(double)(1)))>0.5){
			if(int_src[a].y++>GRIDY)int_src[a].y=GRIDY;
		}
		else{
			if(int_src[a].y--==0)int_src[a].y=0;
		}

	}
	//fft_frame_iris(int_data.fft_iris_ceil,frame);
	return 0;
}


int fft_frame_iris(double rad,pel_t *frame){
	/* now for some FFT fun */
	/* fft_in is our fft in and fft_out is the (imaginary) fft output */
	unsigned GRIDYH=(GRIDY/2)+1,o,a,x,y;

	double fft_in_r[GRIDX*GRIDY], fft_in_g[GRIDX*GRIDY], fft_in_b[GRIDX*GRIDY], fft_in_a[GRIDX*GRIDY];
	double fft_in2_r[GRIDX*GRIDY], fft_in2_g[GRIDX*GRIDY], fft_in2_b[GRIDX*GRIDY], fft_in2_a[GRIDX*GRIDY];
	fftw_complex fft_out_r[GRIDX*GRIDYH], fft_out_g[GRIDX*GRIDYH], fft_out_b[GRIDX*GRIDYH], fft_out_a[GRIDX*GRIDYH];

	fftw_plan p_r2c,p_c2r;

	/* fill fft_in with the colors */
	o=0;a=0;for(x=0;x<GRIDX;x++)for(y=0;y<GRIDY;y++,o+=PEL_NUMEL){
		fft_in_r[a]=frame[o];
		fft_in_g[a]=frame[o+1];
		fft_in_b[a]=frame[o+2];
		fft_in_a[a]=frame[o+3];
		a++;
	}

	/* do the fft */
	p_r2c=fftw_plan_dft_r2c_2d(GRIDX,GRIDY,fft_in_r,fft_out_r,FFTW_ESTIMATE);
	fftw_execute(p_r2c); 
	p_r2c=fftw_plan_dft_r2c_2d(GRIDX,GRIDY,fft_in_g,fft_out_g,FFTW_ESTIMATE);
	fftw_execute(p_r2c); 
	p_r2c=fftw_plan_dft_r2c_2d(GRIDX,GRIDY,fft_in_b,fft_out_b,FFTW_ESTIMATE);
	fftw_execute(p_r2c); 
	p_r2c=fftw_plan_dft_r2c_2d(GRIDX,GRIDY,fft_in_a,fft_out_a,FFTW_ESTIMATE);
	fftw_execute(p_r2c); 

	/* remove some of the high frequencies by taking out the middle bit */
	a=0;
	for(x=0;x<GRIDX;x++){
		for(y=0;y<GRIDYH;y++){
			if(pow(x*x+y*y,0.5)<rad){
				fft_out_r[x*GRIDYH+y][0]=0.0; fft_out_r[x*GRIDYH+y][1]=0.0;
				fft_out_g[x*GRIDYH+y][0]=0.0; fft_out_g[x*GRIDYH+y][1]=0.0;
				fft_out_b[x*GRIDYH+y][0]=0.0; fft_out_b[x*GRIDYH+y][1]=0.0;
				fft_out_a[x*GRIDYH+y][0]=0.0; fft_out_a[x*GRIDYH+y][1]=0.0;
				a++;
			}
		}
	}

	/* now we are free to do the inverse */
	p_c2r=fftw_plan_dft_c2r_2d(GRIDX,GRIDY,fft_out_r,fft_in2_r,FFTW_ESTIMATE);
	fftw_execute(p_c2r);
	p_c2r=fftw_plan_dft_c2r_2d(GRIDX,GRIDY,fft_out_g,fft_in2_g,FFTW_ESTIMATE);
	fftw_execute(p_c2r);
	p_c2r=fftw_plan_dft_c2r_2d(GRIDX,GRIDY,fft_out_b,fft_in2_b,FFTW_ESTIMATE);
	fftw_execute(p_c2r);
	p_c2r=fftw_plan_dft_c2r_2d(GRIDX,GRIDY,fft_out_a,fft_in2_a,FFTW_ESTIMATE);
	fftw_execute(p_c2r);

	o=0;a=0;for(x=0;x<GRIDX;x++)for(y=0;y<GRIDY;y++,o+=PEL_NUMEL){
		//frame[o]=(unsigned)pow(fft_out_r[x*GRIDYH+y][0]*fft_out_r[x*GRIDYH+y][0]+fft_out_r[x*GRIDYH+y][1]*fft_out_r[x*GRIDYH+y][1],0.5);
		//frame[o+1]=(unsigned)pow(fft_out_g[x*GRIDYH+y][0]*fft_out_g[x*GRIDYH+y][0]+fft_out_g[x*GRIDYH+y][1]*fft_out_g[x*GRIDYH+y][1],0.5);
		//jframe[o+2]=(unsigned)pow(fft_out_b[x*GRIDYH+y][0]*fft_out_b[x*GRIDYH+y][0]+fft_out_b[x*GRIDYH+y][1]*fft_out_b[x*GRIDYH+y][1],0.5);
		//frame[o+3]=(unsigned)pow(fft_out_a[x*GRIDYH+y][0]*fft_out_a[x*GRIDYH+y][0]+fft_out_a[x*GRIDYH+y][1]*fft_out_a[x*GRIDYH+y][1],0.5);

		//printf("%d ",(unsigned)pow(fft_out[x][y][0]*fft_out[x][y][0]+fft_out[x][y][1]*fft_out[x][y][1],0.5)/(GRIDX*GRIDY));
		//printf("%f+%fi\n",fft_out_a[x*GRIDYH+y][0],fft_out_a[x*GRIDYH+y][1]);
		
		frame[o]=(unsigned)abs(fft_in2_r[x*GRIDY+y]/(GRIDX*GRIDY));
		frame[o+1]=(unsigned)abs(fft_in2_g[x*GRIDY+y]/(GRIDX*GRIDY));
		frame[o+2]=(unsigned)abs(fft_in2_b[x*GRIDY+y]/(GRIDX*GRIDY));
		frame[o+3]=(unsigned)abs(fft_in2_a[x*GRIDY+y]/(GRIDX*GRIDY));


		//if(frame[o]>200)frame[o]=0;
		//if(frame[o+1]>200)frame[o+1]=0;
		//if(frame[o+2]>200)frame[o+2]=0;
		//if(frame[o+3]>200)frame[o+3]=0;
		//frame[o]=(unsigned)fft_in_r[a]/GRIDX*GRIDY+frame[o];
		//frame[o+1]=(unsigned)fft_in_g[a]/GRIDX*GRIDY+frame[o+1];
		//frame[o+2]=(unsigned)fft_in_b[a]/GRIDX*GRIDY+frame[o+2];
		//frame[o+3]=(unsigned)fft_in_a[a]/GRIDX*GRIDY+frame[o+2];
		a++;
	}

	fftw_destroy_plan(p_r2c); fftw_destroy_plan(p_c2r);
	
	return 0;
}


/* PLUGIN STRUCTURE ******************/
/* These are the structures used by the main thread to find the
 * subroutines.
 */

/* MATRIX PLUGIN: */
plugin_data_t mat_plugin={mat_init,mat_gen,0,&mat_data};

/* FIRE PLUGIN: */
plugin_data_t fire_plugin={fire_init,fire_gen,0,&fire_data};

/* BLOB PLUGIN: */
plugin_data_t blob_plugin={0,blob_gen,0,&blob_data};

/* FLASH PLUGIN: */
plugin_data_t flash_plugin={0,flash_gen,0,&flash_data};

/* TEXT PLUGIN: */
plugin_data_t text_plugin={text_init,text_gen,0,&text_data};

/* DEBUG PLUGIN: */
plugin_data_t debug_plugin={debug_init,debug_gen,0,&debug_data};

/* EVENODD PLUGIN: */
plugin_data_t evenodd_plugin={evenodd_init,evenodd_gen,0,&evenodd_data};

/* INTERFERENCE PLUGIN: */
plugin_data_t interference_plugin={int_init,int_gen,0,&int_data};
