/* lut.c */
/* This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


/* Some subroutines for color LUT manipulation.
 */

/* This defines an LUT structure. Essentially, all this provides is a
 * subroutine (setLUT(...)) that makes for a clean way to initialize
 * certain points in an LUT, and then a subroutine (makeLUT(...)) that will
 * linearly interpolate between the points you provide to make a nice,
 * smooth color change.
 */

/* You need to set LUT_ITER to the number of iterations with which you are
 * dealing. That way, the LUT routines will know not to overflow.
 * LUT's should have LUT_ITER+1 elements.
 */


/* INCLUDES **************************/


#include "lut.h"


/* GLOBALS ***************************/


/* The maximum size of all LUT's: */
unsigned lut_iter=0;


/* CODE ******************************/


/* Just to make the LUT initializer cleaner.
 * Checks for overflow -- if index>lut_iter, it sets lut_iter instead.
 */
void lut_set(lut_t *lut,unsigned index,unsigned a,unsigned r,unsigned g,unsigned b){
  if(index>lut_iter)index=lut_iter;
  lut[index].a=a;lut[index].r=r;lut[index].g=g;lut[index].b=b;lut[index].u=1;
}


/* Fills in the LUT.
 * The first and last elements should be used before this is called,
 * or their values at the end will be unpredictable.
 */
void lut_fill(lut_t *lut){
  unsigned long st,end;
  double A,R,G,B,As,Rs,Gs,Bs;
  lut[0].u=lut[lut_iter].u=1;
  st=0;end=1;while(end<=lut_iter){
    /* Find an appropriate end.
     * (Note: This won't be infinite, because I fixed it 3 lines up.)
		 */
    while(!lut[end].u)++end;
    /* Interpolate between st and end. */
    A=lut[st].a;R=lut[st].r;G=lut[st].g;B=lut[st].b;
    As=(lut[end].a-A)/(end-st);Rs=(lut[end].r-R)/(end-st);
		Gs=(lut[end].g-G)/(end-st);Bs=(lut[end].b-B)/(end-st);
		while(++st<end){
      A+=As;R+=Rs;G+=Gs;B+=Bs;
			lut[st].a=A+.5;lut[st].r=R+.5;lut[st].g=G+.5;lut[st].b=B+.5;
    }
    ++end;
  }
}
