/* plugin.h */

/* A standard include file for disco dance floor plugins. */

/* The format for plugins:
 * All plugins include this file for some standards (like grid size, number
 * of brightness levels, some structures, etc.).  They have three routines
 * called by the main thread, and one data structure that can be changed by
 * the main thread.  These may be named whatever you want, but you have a
 * plug_data_t structure (defined in this file) that points to those.  That
 * way, the main thread can just assemble a list of these structures, from
 * which it can pick plugins at random, or whatever.  The programmer can
 * then add new plugins by simply adding the structure for the plugin to
 * the list.
 * The three routines are:
 * 1) init(seed), which is called when the main thread switches to the
 *    plugin.  The main thread can pass a "random" seed to the routine,
 *    which the routine may ignore if it is not needed.
 * 2) generate(*frame), which actually generates a frame.  This is
 *    stored in the frame pointer, which is four bytes per pel (ARGB), data
 *    arranged by rows (so ARGB, then left to right, then top to bottom).
 *    The data for the green component of the 2nd row from the top, 3rd
 *    column from the left is: frame[1*GRIDLINE + 2*PEL_NUMEL + 2]
 *    (row 1, column 2).  Each byte can range from 0 to GRIDMAX.
 * 3) end(), which is called when the main thread wants to stop running the
 *    plugin.  This is for freeing any memory that the plugin may have
 *    allocated.
 * All routines should return 0 for success, negative for error.
 * To reduce number of parameters passed, the routines should not need to
 * be able to be called by multiple threads, so they may safely store data
 * and allocate memory in pointers given by static variables and global
 * variables.
 */

/* The main thread is intended to call the init() routine, and then use
 * generate() to make a frame, which it will send to the dance floor when
 * the timing is right, and then call end() when it is done with the plugin.
 * To be interesting, generate() should remember its state between calls,
 * so that it can give a different frame each time it is called up until
 * end() is called.  Note that generate() does not have to clear the frame
 * -- it is allowed to use whatever was there beforehand.  Try not to have
 * generate() clear the frame -- this way, some plugins may be combined.
 * Likewise, plugins should not share data.  NOTE: The plugin_data_t
 * structure elements may be NULL -- this means there is no subroutine for
 * that function, or there is not data, so the main thread needs to check
 * for NULL pointers.
 */

#include <inttypes.h>
#ifndef PLUGIN_H
#define PLUGIN_H

#define CHUNK 1024
#define BITS 8
#define CHANNELS 2
#define AUDIO_BUFFSIZE 4092
volatile uint8_t audio_buff[AUDIO_BUFFSIZE];

/* DEFINES ***************************/


/* Number of elements in a pel (ARGB is four): */
#define PEL_NUMEL (4)
/* Floor grid size:  GRIDX cells by GRIDY cells. */
#define GRIDX (16)
#define GRIDY (10)
/* Also, when computing offsets, it may be handy to have GRIDX times
 * PEL_NUMEL and that times GRIDY.
 * The compiler is good enough at optimizing that it should do these
 * multiplies at compiler time:
 */
#define GRIDLINE (GRIDX*PEL_NUMEL)
#define GRIDSIZE (GRIDX*GRIDY*PEL_NUMEL)

/* The max brightness value for a pel. */
#define GRIDMAX (pel_t)(-1)


/* TYPES *****************************/


/* The integer type used for random seeds. */
typedef unsigned rand_t;
/* The data type that one color of one grid pel is stored in. */
typedef unsigned char pel_t;


/* The data structure, containing variables that can be changed to affect
 * how frames are generated.  Specifics of how they are used are left up to
 * the particular plugin.  They do not need to be all used, but a plugin
 * that uses any of them needs to be able to handle any possible value of
 * them.
 */
typedef struct {
	/* Each frame has a fast fourier transform done on it.  This is the lower
	 * end to the FFT radius (for doing low pass filtering */
	double	fft_iris_ceil;
	//double	fft_iris_floor;
	/* Roughly, how dense to make the image. */
	unsigned density;
	/* Affects color. */
	unsigned color;
	/* Affects direction. */
	unsigned dir;
	/* Affects some overall position shift. */
	unsigned shift;
	/* Affects some overall speed. */
	unsigned speed;
	/* Affects some overall phase. */
	unsigned phase;
	/* Some color values to use: */
	pel_t color0[PEL_NUMEL],color1[PEL_NUMEL],color2[PEL_NUMEL],color3[PEL_NUMEL];
	/* A position. */
	int x,y;
} data_t;


/* The structure used to give pointers to the three plugin subroutines
 * called from the main thread.  These are described in the header
 * comments.
 */
typedef struct {
	/* The three functions. */
	int (*init)(rand_t seed);
	int (*generate)(pel_t *frame);
	int (*end)(void);
	/* The data. */
	data_t *data;
} plugin_data_t;


#endif
