/* serial.c */

/* This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/* Three routines.
 * Call serial_open() to open a connection and set it up.
 * Call serial_frame() to write a frame.
 * Call serial_close() to close and release the serial port.
 */

/* Protocol is pretty simple.  Write a 0 byte to signal the first byte of
 * the frame.  Every thing else is two bytes per pixel, with 4 bits per
 * color channel (four channels).  Any 0 byte is incremented to a 1.
 */


/* INCLUDES **************************/


#include <fcntl.h>
#include <termios.h>
#include <poll.h>

#include <errno.h>

#include "plugin.h"

#include <string.h>


/* LOCAL GLOBALS *********************/

/* PIXEL LOOKUP TABLE ****************/
/* You might wonder about this.  Basically, we have many different
 * configurations of tiles with different dimensions and so forth.  Plus,
 * for efficiency we wire them up in some wierd way.  Instead of banging my
 * head against the keyboard for hours, I'm just going to use a lookup
 * table.  This table was generated from an OpenOffice Calc spreadsheet.
 * */

/* 8x8 
const unsigned pixelLUT[64] = {
0,8,16,24,32,40,48,56,
57,49,41,33,25,17,9,1,
2,10,18,26,34,42,50,58,
59,51,43,35,27,19,11,3,
4,12,20,28,36,44,52,60,
61,53,45,37,29,21,13,5,
6,14,22,30,38,46,54,62,
63,55,47,39,31,23,15,7
};
const unsigned pixelLUT[128] = {
48,32,16,0,1,17,33,49,50,34,18,2,3,19,35,51,
52,36,20,4,5,21,37,53,54,38,22,6,7,23,39,55,
56,40,24,8,9,25,41,57,58,42,26,10,11,27,43,59,
60,44,28,12,13,29,45,61,62,46,30,14,15,31,47,63,
79,95,111,127,126,110,94,78,77,93,109,125,124,108,92,76,
75,91,107,123,122,106,90,74,73,89,105,121,120,104,88,72,
71,87,103,119,118,102,86,70,69,85,101,117,116,100,84,68,
67,83,99,115,114,98,82,66,65,81,97,113,112,96,80,64,
};
*/
const unsigned pixelLUT[128] = {
0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,
31,30,29,28,27,26,25,24,23,22,21,20,19,18,17,16,
32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,
63,62,61,60,59,58,57,56,55,54,53,52,51,50,49,48,
64,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,
95,94,93,92,91,90,89,88,87,86,85,84,83,82,81,80,
96,97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,
127,126,125,124,123,122,121,120,119,118,117,116,115,114,113,112
};




/* The filehandle: */
static int serial_hand=-1;


/* The frame buffer: */
#define BUFSIZE (GRIDX*GRIDY*4+1)
static unsigned char serial_buf[BUFSIZE];
/* the backup frame buffer */
static unsigned char serial_bak[BUFSIZE];

/* The special byte set at the beginning of each buffer: */
#define START_BYTE     0xAA
/* The byte to replace every START_BYTE with in the main stream: */
#define START_BYTE_REP 0xAB


/* CODE ******************************/


/* Opens the serial port in non-blocking mode, and sets it up
 * the way I want.  Give it the device name.
 * returns 0 for success, -1 for error (and sets errno).
 */
int serial_open(const char *name){
	struct termios t;
	if(serial_hand>=0){errno=EEXIST;return -1;}
	serial_hand=open(name,O_RDWR|O_NONBLOCK);
	if(serial_hand<0){return -1;}
	/* Start with a raw terminal: */
	cfmakeraw(&t);
	/* The flags I want: */
	t.c_cflag=CS8 | CREAD;
	/* Baud is 2mbps. */
	if(cfsetispeed(&t,B2000000) || cfsetospeed(&t,B2000000)){return -1;}
	if(tcsetattr(serial_hand,TCSANOW,&t)){return -1;}

	/* since we are also using sockets to transmit the values, we should open
	 * a local socket and write our data there (non-blocking as well) */


	return 0;
}


/* Closes the serial output. */
void serial_close(){
	if(serial_hand>=0){close(serial_hand);serial_hand=-1;}
}


/* Writes the given frame (4 bytes per pixel) to the output.
 * First, waits for the last write to finish.
 * Returns 0 for success, -1 for error (sets errno).
 */
int serial_frame(pel_t *frame){
	unsigned a,b,c,x,y;
	int k;
	struct pollfd pfd;
	if(serial_hand<0){errno=EBADF;return -1;}
	/* Wait for writing to be allowed: */
	for(;;){
		pfd.fd=serial_hand;pfd.events=POLLOUT;
		if(poll(&pfd,1,-1)<0){return -1;}
		if(!(pfd.revents&POLLOUT)){continue;}
		/* Check that things look OK: */
		if(write(serial_hand,&pfd,0)<0){
			switch(errno){
				default:return -1;
				case EAGAIN:continue;
			}
		}else{break;}
	}
	/* Make the frame, and write everything.  The handle should be
	 * non-blocking, so this should return immediately once things are ready.
	 */
	serial_buf[0]=START_BYTE;
	for(a=0,b=1;b<BUFSIZE;a+=PEL_NUMEL,b+=4){
		serial_buf[b  ]=frame[a];
		if(serial_buf[b]==START_BYTE)serial_buf[b]=START_BYTE_REP;

		serial_buf[b+1]=frame[a+1];
		if(serial_buf[b+1]==START_BYTE)serial_buf[b+1]=START_BYTE_REP;

		serial_buf[b+2]=frame[a+2];
		if(serial_buf[b+2]==START_BYTE)serial_buf[b+2]=START_BYTE_REP;

		serial_buf[b+3]=frame[a+3];
		if(serial_buf[b+3]==START_BYTE)serial_buf[b+3]=START_BYTE_REP;
	}

	/* re-order serial_buf */
	/* make sure k is a signed integer for this version */
	serial_bak[0]=serial_buf[0];
	c=0;
	for(b=1;b<BUFSIZE;b+=4){
		c=pixelLUT[(b-1)/4]*4+1;
		serial_bak[b ]=serial_buf[c];
		serial_bak[b+1]=serial_buf[c+1];
		serial_bak[b+2]=serial_buf[c+2];
		serial_bak[b+3]=serial_buf[c+3];
	}

	memcpy(serial_buf,serial_bak,sizeof(serial_bak));

	for(;;){
		if(write(serial_hand,serial_buf,BUFSIZE)<0){
			switch(errno){
				default:return -1;
				case EAGAIN:continue;
			}
		}else{break;}
	}
	return 0;
}
